#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <time.h>
#include <gl/glut.h>

#define abs(a)  ((a)<0 ? -(a) : (a))
#define sgn(a)  ((a)<0 ? -1 : 1)
#define max(a,b) ((a)>(b) ? (a):(b))
#define min(a,b)  ((a)<(b) ? (a):(b))

#define MAX_X	666
#define MIN_X	0
#define MAX_Y	444
#define MIN_Y	0

#define MAX_M	22
#define MAX_N	111
#define MAX_PUT 50

typedef struct {
	double x, y, z;
	long xp, yp, zp, i;
	double A, B, C;
} tocka_t;

typedef struct {
	int n;
	int t[MAX_N];
	int vidljiv;
} poligon_t;

int n, m;
tocka_t t[MAX_N];
poligon_t p[MAX_M*MAX_N+2];
tocka_t ociste, glediste, izvor;

int putanja_n;
tocka_t putanja_t[MAX_PUT];

clock_t start;


int faktorijela (int x)
{
	return (x <= 1 ? 1 : x * faktorijela (x-1));
}


double bernstein (int i, int n, double t)
{
	return (double) faktorijela (n) / (faktorijela (i) * faktorijela (n-i)) * pow (t, i) * pow (1-t, n-i);
}


void putanja (double t, double *x, double *y, double *z)
{
	int i;

	if (t >= 0.0 && t <= 1.0) {
		*x = *y = *z = 0;
		for (i=0; i < putanja_n+1; i++) {
			*x += putanja_t[i].x * bernstein (i, putanja_n, t);
			*y += putanja_t[i].y * bernstein (i, putanja_n, t);
			*z += putanja_t[i].z * bernstein (i, putanja_n, t);
		}
	}
}


void projekcijaTocke (tocka_t *t)
{
	tocka_t g1, g2;
	double sina, cosa, sinb, cosb, h;
	tocka_t t1, t2, t3, t4, t5;

	// T1
	t1.x = t->x - ociste.x;
	t1.y = t->y - ociste.y;
	t1.z = t->z - ociste.z;
	g1.x = glediste.x - ociste.x;
	g1.y = glediste.y - ociste.y;
	g1.z = glediste.z - ociste.z;

	// T2
	if (g1.x != 0 || g1.y != 0) {
		sina = (double) g1.y / sqrt (g1.x*g1.x + g1.y*g1.y);
		cosa = (double) g1.x / sqrt (g1.x*g1.x + g1.y*g1.y);
		t2.x = t1.x * cosa + t1.y * sina;
		t2.y = - t1.x * sina + t1.y * cosa;
		t2.z = t1.z;
		g2.x = (double) sqrt (g1.x*g1.x + g1.y*g1.y);
		g2.y = 0;
		g2.z = g1.z;
	} else {
		t2 = t1;
		g2 = g1;
	}

	// T3
	if (g2.x != 0 || g2.y != 0) {
		sinb = (double) g2.x / sqrt (g2.x*g2.x + g2.z*g2.z);
		cosb = (double) g2.z / sqrt (g2.x*g2.x + g2.z*g2.z);
		t3.x = (double) t2.x * cosb - t2.z * sinb;
		t3.y = t2.y;
		t3.z = (double) t2.x * sinb + t2.z * cosb;
	} else {
		t3 = t2;
	}

	// T4
	t4.x = t3.y;
	t4.y = - t3.x;
	t4.z = t3.z;

	// T5
	t5.x = - t4.x; 
	t5.y = t4.y;
	t5.z = t4.z;

	// projekcija
	h = (double) sqrt (pow (ociste.x-glediste.x, 2) + pow (ociste.y-glediste.y, 2) + pow (ociste.z-glediste.z, 2));
	t->xp = (long) ((double) t5.x * h / t5.z) + MAX_X/2;
	t->yp = (long) ((double) t5.y * h / t5.z) + MAX_Y/2;
	t->zp = (long) sqrt (pow (t->x - ociste.x, 2) + pow (t->y - ociste.y, 2) + pow (t->z - ociste.z, 2));
}


void dda (long ys, long ye, long xs, long xe, long *L, long *D)
{
	long y, xi, xf, mi, mf;

	glColor3f (1.0, 0.1, 0.1);
	glBegin (GL_POINTS);

	if (ys == ye) {
		if (ys < MIN_Y || ys > MAX_Y)
			return;
		if (xs < xe)
			L[ys] = xs;
		else
			D[ys] = xs;
	}
	else {
		xi = xs;
		xf = - abs (ye - ys);
		mi = abs ((xe - xs) / (ye - ys));
		mf = 2 * abs ((xe - xs) % (ye - ys));
		for (y = ys; y != ye + sgn (ye - ys); y += sgn (ye - ys)) {
			if (ye > ys) {
				if (y >= MIN_Y && y <= MAX_Y)
					L[y] = xi;
			} else {
				if (y >= MIN_Y && y <= MAX_Y)
					D[y] = xi;
			}
			xi += mi * sgn (xe - xs);
			xf += mf;
			if (xf >= 0) {
				xi += sgn (xe - xs);
				xf -= 2 * abs (ye - ys);
			}
		}
	}
	
	glEnd ();
}


#define V(i,j) t[p[i].t[j]]

long L[MAX_Y], D[MAX_Y];
long Li[MAX_Y], Di[MAX_Y];

void prikazPoligona (int j)
{
	int i, i2, x, y;

	for (y = MIN_Y; y <= MAX_Y; y++)
		L[y] = D[y] = Li[y] = Di[y] = 0;

	for (i=0; i < p[j].n; i++) {
		i2 = (i+1) % p[j].n;
		dda (V(j,i).yp, V(j,i2).yp, V(j,i).xp, V(j,i2).xp, L, D);
		dda (V(j,i).yp, V(j,i2).yp, V(j,i).i, V(j,i2).i, Li, Di);
	}

	for (y = MIN_Y; y <= MAX_Y; y++) {
		if (L[y] || D[y]) {
			glBegin (GL_POINTS);
			for (x=0; x <= D[y]-L[y]; x++) {
				i = Li[y];
				if (x > 0 && Di[y] != Li[y])
					i += (Di[y]-Li[y]) * x / (D[y]-L[y]);
				glColor3f (0.2 + i*0.1/256, 0.3 + i*0.6/256, 0.2 + i*0.1/256);
				glVertex2i (L[y] + x, y);
			}
			glEnd ();
		}
	}
}


void izracunPogleda ()
{
	int i, j, k, f;
	double A, B, C, a, b, c, cosa;

	for (i=0; i < (m+1)*n; i++) {
		projekcijaTocke (&(t[i]));
		t[i].A = t[i].B = t[i].C = 0;
	}

	for (i=0; i < putanja_n; i++) {
		projekcijaTocke (&(putanja_t[i]));
	}

	for (j=0; j < n*m+2; j++) {
		A = (V(j,2).y - V(j,1).y) * (V(j,3).z - V(j,1).z) - (V(j,2).z - V(j,1).z) * (V(j,3).y - V(j,1).y);
		B = - (V(j,2).x - V(j,1).x) * (V(j,3).z - V(j,1).z) + (V(j,2).z - V(j,1).z) * (V(j,3).x - V(j,1).x);
		C = (V(j,2).x - V(j,1).x) * (V(j,3).y - V(j,1).y) - (V(j,2).y - V(j,1).y) * (V(j,3).x - V(j,1).x);
		a = V(j,0).x - ociste.x;
		b = V(j,0).y - ociste.y;
		c = V(j,0).z - ociste.z;

		p[j].vidljiv = (int) ((A*a + B*b + C*c) / (sqrt (a*a + b*b + c*c) * sqrt (A*A + B*B + C*C)) >= 0);

		f = (j > 0 && j < n*m+1 ? 4 : 3);
		for (i=0; i < p[j].n; i++) {
			V(j,i).A += -A / sqrt (A*A + B*B + C*C) / f;
			V(j,i).B += -B / sqrt (A*A + B*B + C*C) / f;
			V(j,i).C += -C / sqrt (A*A + B*B + C*C) / f;
		}
	}

	for (k=0; k<=m; k++) {
		for (i=0; i < n; i++) {
			j = k*n+i;
			a = (izvor.x - t[j].x);
			b = (izvor.y - t[j].y);
			c = (izvor.z - t[j].z);
			if ((a || b || c) && (t[j].A || t[j].B || t[j].C)) {
				cosa = (t[j].A*a + t[j].B*b + t[j].C*c) / (sqrt (a*a + b*b + c*c) * sqrt (pow (t[j].A, 2) + pow (t[j].B, 2) + pow (t[j].C, 2)));
				t[j].i = (int) (20 + max (0, cosa * 200));
			} else {
				t[j].i = 20;
			}
		}
	}
}


void citanjeDatoteke (char *datoteka)
{
	FILE *f;
	int i, j;

	f = fopen (datoteka, "r");
	if (!f) {
		printf ("Datoteka %s ne postoji!?\n", datoteka);
		exit (1);
	}

	if (fscanf (f, "%lf %lf %lf", &(ociste.x), &(ociste.y), &(ociste.z)) != 3
	||  fscanf (f, "%lf %lf %lf", &(glediste.x), &(glediste.y), &(glediste.z)) != 3) {
		printf ("Molimo osigurajte da je datoteka %s u ispravnom formatu.\n", datoteka);
		exit (1);
	}

	fscanf (f, "%d %d", &n, &m);
	for (i=0; i < (m+1)*n  &&  fscanf (f, "%lf %lf %lf", &(t[i].x), &(t[i].y), &(t[i].z)) == 3; i++);

	if (n < 3  ||  m < 1  ||  i < (m+1)*n) {
		printf ("Molim osigurajte da je datoteka u ispravnom formatu.\n", datoteka);
		exit (1);
	}

	for (i = 0; fscanf (f, "%lf %lf %lf", &(putanja_t[i].x), &(putanja_t[i].y), &(putanja_t[i].z)) == 3; i++);
	putanja_n = i-1;
	fclose (f);

	p[0].n = p[m*n+1].n = n;
	for (i=0; i<n; i++) {
		p[0].t[i] = i;
		p[m*n+1].t[n-1-i] = m*n + i;
	}
	for (j=0; j<m; j++) {
		for (i=0; i<n; i++) {
			p[j*n+i+1].n = 4;
			p[j*n+i+1].t[0] = j*n + (i+1) % n;
			p[j*n+i+1].t[1] = j*n + i;
			p[j*n+i+1].t[2] = (j+1)*n + i;
			p[j*n+i+1].t[3] = (j+1)*n + (i+1) % n;
		}
	}
}


void prikazKrivulje ()
{
	double t;
	tocka_t tocka;

	glColor3f (0.2, 0.2, 1.0);
	glBegin (GL_LINE_LOOP);
	for (t=0; t<=1; t+=0.1) {
		putanja (t, &tocka.x, &tocka.y, &tocka.z);
		projekcijaTocke (&tocka);
		glVertex2i (tocka.xp, tocka.yp);
	}
	glEnd ();
}


void obradaPrikaza (void)
{
	int i;

	glClear (GL_COLOR_BUFFER_BIT);

	for (i=0; i < m*n+2; i++)
		if (p[i].vidljiv)
			prikazPoligona (i);

	glutSwapBuffers ();
}


void obradaVremena (void)
{
	static double t = 0;
	long int d, D;

	t = max (0, min ((double) (clock() - start) / 5000, 1));

	if (t <= 1) {
		putanja (t, &ociste.x, &ociste.y, &ociste.z);

/*		ociste.x = (700 + 6000*t) * cos (2 * 3.14 * t);
		ociste.y = (700 + 6000*t) * sin (2 * 3.14 * t);
		ociste.z = 270;
*/
		D = sqrt (pow (ociste.x - 0, 2) + pow (ociste.y - 0, 2) + pow (ociste.z - 100, 2));
		d = max (0, D - 500);

		glediste.x = 0 + (ociste.x - 0) * d / D;
		glediste.y = 0 + (ociste.y - 0) * d / D;
		glediste.z = 100 + (ociste.z - 100) * d / D;

		izracunPogleda ();
		glutPostRedisplay ();
		t += 1.0/500;
		//glutTimerFunc (1000/30, obradaVremena, id);
	}
}


void obradaTipkovnice (unsigned char tipka, int x, int y)
{
	start = clock ();
}


void main (int argc, char *argv[])
{
	char *datoteka = "ulaz.txt";

	if (argc == 2)
		datoteka = argv[1];
	citanjeDatoteke (datoteka);

	izvor.x = -400;
	izvor.y = 400;
	izvor.z = 1200;
	putanja (0, &ociste.x, &ociste.y, &ociste.z);
	izracunPogleda ();

	glutInit (&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE);
	glutInitWindowPosition (200, 100);
	glutInitWindowSize (MAX_X, MAX_Y);
	glutCreateWindow ("Laboratorijska vjezba 8");

	glOrtho (0, MAX_X, 0, MAX_Y, 0, 1); 
	glClearColor (1.0, 1.0, 1.0, 1.0);
	glColor3f (0.4, 0.4, 0.4);

	glutDisplayFunc (obradaPrikaza);
	glutKeyboardFunc (obradaTipkovnice);
	//glutTimerFunc (1000, obradaVremena, 0);
	glutIdleFunc (obradaVremena);

	start = clock () + 100;
	glutMainLoop ();
}